from prompt_toolkit.validation import ValidationError


def validate(value, schema):
    # ctrl-d gives us a None
    if value is None:
        return False, None

    vartype = schema.get("type", "string")
    if vartype == "string":
        value = str(value)

    return False, value



import sys

from prompt_toolkit.validation import ValidationError


def validate(value, schema):
    allowed = schema.get("allowed")
    if allowed is None:
        return False, None

    if value not in allowed:
        allowed = ", ".join(allowed)
        raise ValidationError(message=f"Answer must be within [{allowed}]")

    return False, value



"""
Jinja2 extensions.

Source: https://github.com/cookiecutter/cookiecutter
"""
# pylint: disable=abstract-method
import json
import string
import uuid
from secrets import choice

from jinja2.ext import Extension
from slugify import slugify as pyslugify


class JsonifyExtension(Extension):
    """
    Jinja2 extension to convert a Python object to JSON.
    """

    def __init__(self, environment):
        """
        Initialize the extension with the given environment.
        """
        super().__init__(environment)

        def jsonify(obj):
            return json.dumps(obj, sort_keys=True, indent=4)

        environment.filters["jsonify"] = jsonify


class RandomStringExtension(Extension):
    """
    Jinja2 extension to create a random string.
    """

    def __init__(self, environment):
        """
        Jinja2 Extension Constructor.
        """
        super().__init__(environment)

        def random_ascii_string(length, punctuation=False):
            if punctuation:
                corpus = "".join((string.ascii_letters, string.punctuation))
            else:
                corpus = string.ascii_letters
            return "".join(choice(corpus) for _ in range(length))

        environment.globals.update(random_ascii_string=random_ascii_string)


class SlugifyExtension(Extension):
    """
    Jinja2 Extension to slugify string.
    """

    def __init__(self, environment):
        """
        Jinja2 Extension constructor.
        """
        super().__init__(environment)

        def slugify(value, **kwargs):
            """
            Slugifies the value.
            """
            return pyslugify(value, **kwargs)

        environment.filters["slugify"] = slugify


class UUIDExtension(Extension):
    """
    Jinja2 Extension to generate uuid4 string.
    """

    def __init__(self, environment):
        """
        Jinja2 Extension constructor.
        """
        super().__init__(environment)

        def uuid4():
            """
            Generate UUID4.
            """
            return str(uuid.uuid4())

        environment.globals.update(uuid4=uuid4)



from prompt_toolkit.validation import ValidationError


def validate(value, schema):
    # ctrl-d gives us a None
    if value is not None:
        return False, None

    nullable = schema.get("nullable", False)
    if nullable:
        return True, ""
    raise ValidationError(message="Answer is required")



"""
Ask questions and answer them by prompting the user
"""
import sys

from prompt_toolkit import prompt as user_input
from prompt_toolkit.validation import ValidationError, Validator

from . import utils
from .environment import StrictNativeEnvironment
from .validators import allowed, boolean, default, max_length, min_length, nullable, vartype


def validate_answer(value, schema):
    """
    Validate the answer and return modified answer if needed.
    """
    # Order matters. By checking nullable and default first
    # we can make assumptions on values (null? not null? etc.)
    validators = [
        vartype,
        nullable,
        default,
        vartype,  # force the default value type to be of type X
        allowed,
        boolean,
        max_length,
        min_length,
    ]

    # Returns (True, Any) if we are to stop iterating,
    # Returns (Any, Value) if we are to replace value by
    # a new value
    for validator in validators:
        valid, new_value = validator.validate(value, schema)
        if new_value is not None:
            value = new_value
        if valid is True:
            return valid, value
    return True, value


def stdin_input(prompt):
    """
    Convenience function in order to be mocked
    """
    return input(prompt)


def prompt_question(question):
    """
    Ask the question until it is answered or canceled
    """
    # Get question details
    name = question["name"]
    prompt = question["prompt"] + ": "
    description = question.get("description")
    schema = question.get("schema", {})

    if schema.get("hidden", False):
        return name, schema.get("default")

    def prevalidate_answer(x, schema):
        return validate_answer(x, schema)[0]

    validator = Validator.from_callable(lambda x: prevalidate_answer(x, schema))

    while True:
        try:
            if sys.stdin.isatty():
                answer = user_input(
                    prompt,
                    validator=validator,
                    bottom_toolbar=description,
                    validate_while_typing=False,
                )
            else:
                answer = stdin_input(prompt)
            _, answer = validate_answer(answer, schema)
            return name, answer

        except EOFError:
            # ctrl-d was used
            try:
                _, answer = validate_answer(None, schema)
                return name, answer
            except ValidationError:
                continue


def prepare_question(question, answers, env, ctx):
    """
    Determine if the question is to be asked, and
    if so, build a prompt
    Return true if question is to be asked.
    """

    def templatize(what):
        if isinstance(what, str):
            return env.from_string(what).render(**ctx)
        return what

    # Get question details
    name = question["name"]
    schema = question.get("schema", {})

    # Will we prompt this question?
    when = question.get("if")
    when = templatize(when)
    if when is not None:
        question["if"] = when
        if not when:
            return False

    # Prepare answer to prepare default
    answer = (answers or {}).get(question["name"])
    if answer is not None:
        schema["default"] = answer

    # Build default
    # pylint: disable=redefined-outer-name
    default = schema.get("default")
    default = templatize(default)
    if default is not None:
        schema["default"] = default

    # Build prompt
    question["prompt"] = name
    if default is not None:
        question["prompt"] += f" [{default}]"
    return True


def prompt_questions(args, data):
    """
    For every question in the input file, ask the question
    and record the answer in the context

    Fills in `questions`
    """

    # jinja context, everything of ours is "scaffold".* based
    ctx = {"scaffold": {}}

    # Get user options for jinja from questions file.
    jinja2 = data.get("jinja2", {})

    env = StrictNativeEnvironment(context=data.get("extensions"), path=args["template_dir"], **jinja2)

    questions = data.get("questions") or []
    answers = data.get("answers") or {}
    for question in questions:
        if not prepare_question(question, answers, env, ctx):
            continue

        name, answer = prompt_question(question)
        if name is not None:
            question["value"] = answer
            if question.get("schema", {}).get("type", "string") == "boolean":
                answer = utils.string_as_bool(answer)
            ctx["scaffold"][name] = answer
    return env, ctx



import os
import shutil
import sys

import jinja2
from boltons import fileutils

from .. import answers, questions, template
from ..environment import StrictNativeEnvironment


def clone_tree(args, data, ctx):
    # TODO: Verify before questions
    #
    # Get user options
    jinja_vars = data.get("jinja2", {})
    env = StrictNativeEnvironment(context=data.get("extensions"), path=args["template_dir"], **jinja_vars)

    # Get the template source directory (from command line)
    # It's the folder containing another folder named "template"
    source_path = args["template"]
    source_path = os.path.join(source_path, "template")
    source_path = os.path.abspath(source_path)

    if not os.path.isdir(source_path):
        sys.exit('error: no "template" named folder found in source template directory. No templating can be done')

    # For every source file, replace templates, and create file
    output = args.get("output_dir", ".")
    for src in fileutils.iter_find_files(source_path, "*", include_dirs=True):
        dst = os.path.relpath(src, source_path)
        try:
            dst = env.from_string(dst).render(**ctx)
        except jinja2.exceptions.UndefinedError as err:
            sys.exit(f'error: templating error during "{src}" file: {err}')
        dst = os.path.join(output, dst)

        if os.path.islink(src):
            # TODO: Change destination
            assert False, "change link destination"
        elif os.path.isdir(src):
            fileutils.mkdir_p(dst)
        else:
            fileutils.mkdir_p(os.path.dirname(dst))
            try:
                with open(src, encoding="UTF-8") as fd:
                    dst_contents = fd.read()
                    dst_contents = env.from_string(dst_contents).render(**ctx)
                with open(dst, mode="w", encoding="UTF-8") as fd:
                    fd.write(str(dst_contents))

            except UnicodeDecodeError:
                with open(src, mode="rb") as fd:
                    dst_contents = fd.read()
                with open(dst, mode="wb") as fd:
                    fd.write(dst_contents)

            except jinja2.exceptions.UndefinedError as err:
                sys.exit(f'error: while rendering "{src}", encountered template error: {err}')
            except jinja2.exceptions.TemplateSyntaxError as err:
                sys.exit(f'error: while rendering "{src}", encountered template error: {err}')

    return ctx


def clean_tree(args, env, ctx, files):
    if not files:
        return

    # Prepare path
    output_dir = args["output_dir"]
    template_path = os.path.join(os.getcwd(), output_dir)
    template_path = os.path.abspath(template_path)

    for file in files:
        path = file["path"]
        path = env.from_string(path).render(**ctx)

        when = file["if"]
        when = env.from_string(when).render(**ctx)
        if not when:
            continue

        path = os.path.abspath(os.path.join(template_path, path))
        if not path.startswith(template_path):
            sys.exit("error: can only delete files in copy's subpath")

        action = file["action"]
        if action == "remove":
            if os.path.islink(path):
                os.remove(path)
            elif os.path.isfile(path):
                os.remove(path)
            elif os.path.isdir(path):
                shutil.rmtree(path)
        else:
            sys.exit(f'error: unknown "action" for file path "{path}"')


def clone(args):
    """
    Ask the questions, and clone the tree, not a repository
    """
    # Get local folder from which we can load files from
    data = template.get(args)
    data = answers.load(args, data)
    env, ctx = questions.prompt_questions(args, data)

    ctx = clone_tree(args, data, ctx)

    bus = data.get("files") or []
    clean_tree(args, env, ctx, bus)



import sys

import yaml

from ..constants import TEMPLATE_NAME
from . import directory, git


def get(args):
    """
    Find, fetch the template file and return it's contents
    """
    template_path = args["template"]
    for fn in [directory.find, git.find]:
        file, path = fn(template_path)
        if file and path:
            args["template_file"] = file
            args["template_dir"] = path
            break
    else:
        sys.exit(f'error: no "{TEMPLATE_NAME}" file found')

    with open(file, encoding="UTF-8") as fd:
        return yaml.safe_load(fd)



import os

from ..constants import TEMPLATE_NAME


def find(path):
    """
    Search given path, then path/file.yml
    """
    for file in path, os.path.join(path, TEMPLATE_NAME):
        if os.path.exists(file) and not os.path.isdir(file):
            return file, os.path.dirname(file)
    return False, False



"""
Scheme validation of scaffold.yml file
"""

import json
import os
import sys

import cerberus
import yaml

from .. import template

# https://docs.python-cerberus.org/en/stable/validation-rules.html
SCHEMA = r"""
---
extensions:
  type: list
  default: []
  schema:
    type: string

jinja2:
  type: dict
  default: {}
  schema:
    lstrip_blocks:
      type: boolean
    trim_blocks:
      type: boolean

files:
  type: list
  default: []
  schema:
    type: dict
    default: {}
    nullable: true
    schema:
      path:
        type: string
        required: true
      if:
        type: [string, boolean]
      action:
        type: string
        required: true

questions:
  type: list
  default: []
  schema:
    type: dict
    default: {}
    nullable: true
    schema:
      name:
        type: string
        regex: '^\S+$'
        required: true
      description:
        type: string
      value:
        type: [string, boolean]
      if:
        type: [string, boolean]
      schema:
        type: dict
        default: {}
        nullable: true
        schema:
          type:
            type: string
            allowed: ["string", "boolean"]
          default:
            type: [string, boolean]
          hidden:
            type: boolean
            default: false
          nullable:
            type: boolean
            default: true
          min_length:
            type: integer
            min: 0
          max_length:
            type: integer
            min: 0
          allowed:
            type: list
            schema:
              type: string
"""


def _validate_schema(args):
    """
    Validate the YAML part of the template
    """
    data = template.get(args)
    schema = yaml.safe_load(SCHEMA)
    validator = cerberus.Validator(schema)
    if not validator.validate(data):
        sys.exit("error: YAML schema validation error. Location:\n" + str(json.dumps(validator.errors, indent=2)))
    return data


def _validate_template_existance(args):
    """
    Validate that a folder named 'template' exists
    """
    path = args["template_dir"]
    path = os.path.join(path, "template")
    if not os.path.isdir(path):
        sys.exit('error: no "template" directory found in source template location')


def _validate_file_actions(data):
    for file in data.get("files", []):
        action = file.get("action")
        if action not in ["remove"]:
            sys.exit(f'error: unknown file action "{action}"')


def _validate_questions(data):
    for question in data.get("questions", []):
        schema = question.get("schema")
        if not schema:
            continue

        if "allowed" in schema:
            if schema.get("type", "string") != "string":
                sys.exit('error: field "allowed" only allows using a "string" type')
            if "min_length" in schema:
                print('warning: presence of field "allowed" with a "min_length" type')
            if "max_length" in schema:
                print('warning: presence of field "allowed" with a "max_length" type')

        if schema.get("hidden", False):
            if schema.get("default") is None:
                sys.exit('error: field "hidden" requires a "default" value')


def validate(args):
    """
    Called from cli
    """
    data = _validate_schema(args)
    _validate_template_existance(args)
    _validate_questions(data)
    _validate_file_actions(args)
